/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.tic;

import java.util.ArrayList;
import java.util.List;

import org.junit.Test;

import de.bsvrz.kex.lmstic3.vew.daten.tic.KoordinatenReduktion;
import de.bsvrz.kex.lmstic3.vew.daten.tic.elemente.TIC3Koordinaten;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Punkt;

/**
 * Testcase Koordinatenreduktion.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */

public class TestKoordinatenReduktion {

	private double[][] liste = { { 49.63495, 8.45151 }, { 49.63514, 8.45138 }, { 49.63525, 8.45132 },
			{ 49.63537, 8.45128 }, { 49.63548, 8.45126 } };

	/*
	 * <Coordinate name="Koordinate" description="49,63495° 8,45151°"> <Latitude
	 * name="Breitengrad">49.63495</Latitude> <Longitude
	 * name="Längengrad">8.45151</Longitude> </Coordinate> <Coordinate
	 * name="Koordinate" description="49,63514° 8,45138°"> <Latitude
	 * name="Breitengrad">49.63514</Latitude> <Longitude
	 * name="Längengrad">8.45138</Longitude> </Coordinate> <Coordinate
	 * name="Koordinate" description="49,63525° 8,45132°"> <Latitude
	 * name="Breitengrad">49.63525</Latitude> <Longitude
	 * name="Längengrad">8.45132</Longitude> </Coordinate> <Coordinate
	 * name="Koordinate" description="49,63537° 8,45128°"> <Latitude
	 * name="Breitengrad">49.63537</Latitude> <Longitude
	 * name="Längengrad">8.45128</Longitude> </Coordinate> <Coordinate
	 * name="Koordinate" description="49,63548° 8,45126°"> <Latitude
	 * name="Breitengrad">49.63548</Latitude> <Longitude
	 * name="Längengrad">8.45126</Longitude> </Coordinate>
	 */

	/**
	 * Test method for
	 * {@link de.bsvrz.kex.lmstic3.versand.kodierer.Meldungskodierer#run()}.
	 */
	@Test
	public void testReduktion() {
		final List<TIC3Koordinaten> koordinaten = new ArrayList<>();

		for (int i = 0; i < liste.length; i++) {
			// System.out.println("Breite: " + liste[i][0] + ", Länge: " +
			// liste[i][1]);
			final TIC3Koordinaten k = new TIC3Koordinaten();
			k.setBreite(liste[i][0]);
			k.setLaenge(liste[i][1]);
			koordinaten.add(k);
		}

		System.out.println("Eingangsdaten");
		printAbstaende(koordinaten);

		final List<TIC3Koordinaten> redkoordinaten = KoordinatenReduktion.reduziere(koordinaten, 100);

		System.out.println("\nReduzierte Daten");
		printAbstaende(redkoordinaten);
	}

	private void printAbstaende(final List<TIC3Koordinaten> kliste) {

		for (int i = 0; i < kliste.size() - 1; i++) {
			final TIC3Koordinaten k1 = kliste.get(i);
			final TIC3Koordinaten k2 = kliste.get(i + 1);

			final WGS84Punkt p1 = new WGS84Punkt(k1.getLaenge(), k1.getBreite());
			final WGS84Punkt p2 = new WGS84Punkt(k2.getLaenge(), k2.getBreite());

			System.out.println("Punkt: " + p1 + ", " + p2 + ", Abstand " + WGS84Punkt.abstand(p1, p2) + "m");
		}
	}
}
