/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.versand.kodierer;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.lmstic3.versand.ftpclient.LMSFTPClient;
import de.bsvrz.kex.lmstic3.vew.LMSFTPListener;
import de.bsvrz.kex.lmstic3.vew.LMSKommunikationsstatus;
import de.bsvrz.kex.lmstic3.vew.LMSParameter;
import de.bsvrz.kex.lmstic3.vew.LMSProperties;
import de.bsvrz.kex.lmstic3.vew.LMSTIC3;
import de.bsvrz.kex.lmstic3.vew.LMSTIC3LogTools;
import de.bsvrz.kex.lmstic3.vew.LMTIC3Betriebsmeldungen;
import de.bsvrz.kex.lmstic3.vew.LMTIC3Betriebsmeldungen.LMTIC3Betriebsmeldung;
import de.bsvrz.kex.lmstic3.vew.LmsMsg;
import de.bsvrz.kex.lmstic3.vew.MeldungsStatus;
import de.bsvrz.kex.lmstic3.vew.TMCMeldung;
import de.bsvrz.kex.lmstic3.vew.XMLTools;
import de.bsvrz.kex.lmstic3.vew.daten.DavDatenModell;
import de.bsvrz.kex.lmstic3.vew.daten.ExterneMeldungsDaten;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3Converter;
import de.bsvrz.kex.lmstic3.vew.daten.tic.TIC3Exception;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modul Meldungskodierer.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: Meldungskodierer.java 64047 2016-10-27 15:42:01Z gieseler $
 */

public class Meldungskodierer extends Thread
		implements ClientReceiverInterface, ClientSenderInterface, LMSFTPListener, InvalidationListener {

	/**
	 * Listener f&uuml;r &Auml;nderungen der Menge der TMC-Meldungen.
	 */
	class MengeListener implements MutableSetChangeListener {

		@Override
		public void update(final MutableSet set, final SystemObject[] addedObjects,
				final SystemObject[] removedObjects) {
			for (final SystemObject so : addedObjects) {
				debug.info("Die neue Meldung '" + so.getPid() + "' wird hinzugefügt!");
				try {
					hinzufuegenMeldung(so);
				} catch (final TIC3Exception e) {
					debug.error(e.getMessage());
				}
			}

			for (final SystemObject so : removedObjects) {
				debug.info("Die Meldung '" + so.getPid() + "' wird aus der internen Liste entfernt!");
				entfernenMeldung(so);
			}
		}
	}

	/**
	 * Attributgruppe der Kommunikationsparameter.
	 */
	private static final String ATTRIBUTGRUPPE_PARAMETER = "atg.lmsKommunikationsParameter";

	/**
	 * Wurzelelement der TIC-Info Dateien.
	 */
	private static final String XML_ELEMENT_TIC_ROOT = "TIC";

	/**
	 * DebugLogger für Debug-Ausgaben.
	 */
	private final Debug debug = Debug.getLogger();

	/**
	 * DaV Interface.
	 */
	private ClientDavInterface dav;

	/**
	 * Datenmodell der Konfiguration.
	 */
	private DataModel config;

	/**
	 * Objektreferenz auf das Objekt vom Typ "typ.landesMeldeStelle".
	 */
	private ConfigurationObject lms;

	/**
	 * Die Liste der zu versendenden Meldungen.
	 */
	private Map<String, TMCMeldung> meldungen = new HashMap<>();

	/**
	 * FTP-Client Thread.
	 */
	private LMSFTPClient ftpclient;

	/** Datenbeschreibung, mit der Parameter empfangen werden. */
	private DataDescription paramEmpfang;

	/** Datenbeschreibung, mit der Parameter versendet werden. */
	private DataDescription paramVersand;

	/** Parameter. */
	private final LMSParameter lmsParameter = new LMSParameter();

	/** Der Kommunikationsstatus mit der Landesmeldestelle. */
	private LMSKommunikationsstatus kommunikationsStatus;

	/** Lock-Objekt. */
	private final Object runLock = new Object();

	private boolean datenAngemeldet;

	/** Run-Flag. */
	private boolean run = true;

	private boolean parameterEmpfangen;

	private long letzterVersand;

	/**
	 * Konstruktor.
	 *
	 * @param dav
	 *            Datenverteilerverbindung
	 * @param lms
	 *            das die Landesmeldestelle repr&auml;sentierende Systemobjekt
	 * @param mitValidierung
	 *            legt fest, ob die empfangenen Dateien gegen das Schema
	 *            validiert werden sollen
	 */
	public Meldungskodierer(final ClientDavInterface dav, final ConfigurationObject lms, final boolean mitValidierung) {
		super();
		this.dav = dav;
		this.lms = lms;
		config = dav.getDataModel();
		this.setName(this.getClass().getSimpleName());

		DavDatenModell.getInstanz().setVerbindung((ClientDavConnection) dav);

		// selbst starten
		start();
	}

	/**
	 * Bildet den Dateinamen f&uuml;r eine zu versendende Datei.
	 *
	 * @return Dateiname
	 */
	private String bildeDateinamen() {
		return lmsParameter.getVersandDateiNamePrefix() + "_" + System.currentTimeMillis() + ".xml";
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		debug.finest("dataRequest");
	}

	/**
	 * Meldet alle ben&ouml;tigten Daten am Datenverteiler an.
	 */
	private void datenAnmeldenDaV() {
		final DataDescription dataDescription = null;
		SystemObject[] objekte = null;
		AttributeGroup attributeGroup = null;
		Aspect aspect = null;

		if (datenAngemeldet) {
			return;
		}

		// Anmeldungen zum Senden und Empfangen.
		try {
			//
			// Anmeldungen zum Empfang
			//
			// Kommunikationsparameter
			objekte = new SystemObject[] { lms };
			attributeGroup = config.getAttributeGroup(Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER);
			aspect = config.getAspect("asp.parameterSoll");
			paramEmpfang = new DataDescription(attributeGroup, aspect);

			// aktuelle Parameter lesen
			final ResultData resultdaten = dav.getData(lms, paramEmpfang, 0);
			lmsParameter.aktualisieren(resultdaten.getData());

			debug.finer("Anmeldung zum Empfang von Daten beim DaV als SENKE :" + "(Objekt: " + lms.getName()
					+ ") (ATG: " + Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER + ") (ASP: " + "asp.parameterSoll"
					+ ") (OT: " + lms.getType().getName() + ")");

			dav.subscribeReceiver(this, objekte, paramEmpfang, new ReceiveOptions(false), // keine
					// Deltas, keine nachgelieferten Daten
					ReceiverRole.receiver());

			// Menge TMC-Meldungen anmelden
			final MutableSet meldungsmenge = lms.getMutableSet(LMSTIC3.MENGE_TMC_MELDUNGEN);
			debug.info("Menge '" + LMSTIC3.MENGE_TMC_MELDUNGEN + "' am Objekt " + lms + ": " + meldungsmenge);

			// meldungsmenge.addChangeListener(new MengeListener());
			for (final SystemObject meldung : meldungsmenge.getElements()) {
				debug.info("Die aktuelle TMC-Meldung " + meldung + " wird initialisiert");
				try {
					hinzufuegenMeldung(meldung);
				} catch (final TIC3Exception e) {
					debug.error(e.getMessage());
				}
			}
			meldungsmenge.addChangeListener(new MengeListener());

			//
			// Anmeldungen zum Senden
			//
			// Kommunikationsparameter
			objekte = new SystemObject[] { lms };
			attributeGroup = config.getAttributeGroup(Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER);
			aspect = config.getAspect("asp.parameterVorgabe");
			paramVersand = new DataDescription(attributeGroup, aspect);

			debug.finer("Anmeldung zum Senden von Daten an den DaV als SENDER :" + "(Objekt: " + lms.getName()
					+ ") (ATG: " + Meldungskodierer.ATTRIBUTGRUPPE_PARAMETER + ") (ASP: " + "asp.parameterVorgabe"
					+ ") (OT: " + lms.getType().getName() + ")");

			dav.subscribeSender(this, objekte, paramVersand, SenderRole.sender());

			// IvalidationListener
			final DynamicObjectType typ = (DynamicObjectType) config.getType(LMSTIC3.TYP_TMC_MELDUNGEN);
			// typ.addObjectCreationListener(privateListener);
			typ.addInvalidationListener(this);
			datenAngemeldet = true;
		} catch (final OneSubscriptionPerSendData e) {
			debug.error("Daten wurden bereits als Quelle angemeldet: " + dataDescription);
			throw new RuntimeException(e);
		} catch (final DataNotSubscribedException e) {
			// TBD: Die Exception ist wiedersprüchlich dokumentiert --> Roland
			debug.error("Schreibanmeldung fehlgeschlagen: " + dataDescription);
			System.exit(0);
			throw new RuntimeException(e);
		}
	}

	@Override
	public void empfangen(final List<File> dateien) {
		// das sollte hier nicht auftreten
		debug.error("unerwartetes Listener-Event 'empfangen'");
	}

	/**
	 * Entfernt eine Meldung aus der Meldungsliste.
	 *
	 * @param so
	 *            zugeordnetes Systemobjekt
	 */
	void entfernenMeldung(final SystemObject so) {
		synchronized (meldungen) {
			final TMCMeldung rdsMeldung = meldungen.get(so.getPid());
			if (rdsMeldung != null) {
				try {
					rdsMeldung.abmeldenDavVersand(dav);
				} catch (final TIC3Exception e) {
					debug.error("abmeldenDavVersand: " + e.getMessage());
				}
				meldungen.remove(so.getPid());
			}
		}
	}

	/**
	 * F&uuml;get eine Meldung zur Meldungsliste hinzu.
	 *
	 * @param objekt
	 *            Systemobjekt der Meldung
	 * @throws TIC3Exception
	 *             bei Ausnahmen
	 */
	public void hinzufuegenMeldung(final SystemObject objekt) throws TIC3Exception {
		synchronized (meldungen) {
			final TMCMeldung meldung = new TMCMeldung(objekt);
			meldung.anmeldenDavVersand(dav);
			meldungen.put(meldung.getPID(), meldung);
		}
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		debug.finest("isRequestSupported");
		return false;
	}

	@Override
	public void kommunikationsStatus(final LMSKommunikationsstatus status) {
		final boolean changed = kommunikationsStatus != status;

		kommunikationsStatus = status;

		if (kommunikationsStatus == LMSKommunikationsstatus.GESTOERT) {
			debug.fine("Die Meldungen konnten nicht versendet werden");
		}

		if (changed) {
			if (kommunikationsStatus == LMSKommunikationsstatus.GESTOERT) {
				LMSTIC3LogTools.log(debug, LmsMsg.WarningStoerungAusgehend, lms.getName());
				LMTIC3Betriebsmeldungen.sendeMeldung(LMTIC3Betriebsmeldung.KOMMUNIKATION_AUSGEHEND_GESTOERT);
			}
			if (kommunikationsStatus == LMSKommunikationsstatus.NICHT_GESTOERT) {
				LMSTIC3LogTools.log(debug, LmsMsg.InfoStatusAusgehendOK, lms.getName());
				LMTIC3Betriebsmeldungen.sendeMeldung(LMTIC3Betriebsmeldung.KOMMUNIKATION_AUSGEHEND_OK);
			}
		}
	}

	/**
	 * Aktualisiert die laufende Nummer der Kommunikationsparameter.
	 */
	public void laufendeNummerAktualisieren() {
		if (lmsParameter.getAktuelleLaufendeNummer() == lmsParameter.getLaufendeNummer()) {
			return;
		}

		try {
			ResultData resultdaten = dav.getData(lms, paramEmpfang, 0);
			final Data neueDaten = resultdaten.getData().createModifiableCopy();
			neueDaten.getUnscaledValue("LaufendeNummer").set(lmsParameter.getAktuelleLaufendeNummer());
			resultdaten = new ResultData(lms, paramVersand, System.currentTimeMillis(), neueDaten);
			lmsParameter.resetLaufendeNummer();
			dav.sendData(resultdaten);

			// warte bis die Parameter aktualisiert sind
			synchronized (this) {
				try {
					this.wait(1000);
				} catch (final InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		} catch (final DataNotSubscribedException e) {
			LMSTIC3LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
		} catch (final SendSubscriptionNotConfirmed e) {
			LMSTIC3LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
		}
	}

	@Override
	public void run() {

		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
			@Override
			public void uncaughtException(final Thread t, final Throwable e) {
				debug.error("Meldungskodierer beendet (" + e + ")"); //$NON-NLS-1$
				debug.error(LoggerTools.getStackTrace(e));
				run = false;
				if ((ftpclient != null) && ftpclient.isAlive()) {
					ftpclient.stopp();
					try {
						ftpclient.join();
					} catch (final InterruptedException e1) {
					}
				}
			}
		});

		synchronized (this) {
			try {
				datenAnmeldenDaV();

				parameterEmpfangen = false;

				while (run) {
					// auf Parameter warten
					if (!parameterEmpfangen) {
						LMSTIC3LogTools.log(debug, LmsMsg.InfoWarteParameter);
						try {
							wait();
						} catch (final InterruptedException e1) {
						}
					}

					parameterEmpfangen = false;

					if (!isVersandAktiv()) {
						LMSTIC3LogTools.log(debug, LmsMsg.InfoKeinVersand);
						continue;
					}

					ftpclient = new LMSFTPClient(lmsParameter, true);
					ftpclient.addListener(this);
					while (run) {

						wait(1000);
						if (parameterEmpfangen) {
							break;
						}
						try {
							ueberpruefeMeldungen();
						} catch (final TIC3Exception e) {
							debug.error(e.getMessage());
						}
					}

					if (ftpclient != null) {
						ftpclient.stopp();
					}
				}
			} catch (final InterruptedException e) {
				/** */
			}
		}
	}

	/**
	 * Stoppt den Thread.
	 */
	public void stopp() {
		this.run = false;
		this.interrupt();
	}

	/**
	 * &Uuml;berpro&uuml;ft alle Meldungen auf einen notwendigen Versand.
	 *
	 * @throws TIC3Exception
	 *             bei Ausnahmen
	 */
	private void ueberpruefeMeldungen() throws TIC3Exception {
		final List<TMCMeldung> versenden = new ArrayList<>();

		synchronized (meldungen) {
			boolean spontanVersand = false;

			for (final TMCMeldung meldung : meldungen.values()) {
				try {
					if (!meldung.isDavDatenOK() || (meldung.getStatus() == MeldungsStatus.FEHLERHAFT)) {
						continue;
					}
					if (!meldung.istZeitGueltig()) {
						debug.info("Die TMC-Meldung: " + meldung
								+ " ist nicht (mehr) gültig (EndeZeit<aktuelle Zeit) und wird deshalb nicht versendet");

						// (erstmal) nicht mehr bearbeiten
						meldung.setStatus(MeldungsStatus.FEHLERHAFT);
						continue;
					}

					if (meldung.getStatus() == MeldungsStatus.AKTUALISIERT) {
						// Zustandsänderung nach Versand
						if (meldung.isVersendet()) {
							meldung.publiziereVersendet(dav);
							continue;
						}

						spontanVersand = true;
						meldung.setStatus(MeldungsStatus.GUELTIG);
					}
					versenden.add(meldung);
				} catch (final Exception e) {
					debug.error(e.getMessage());
				}
			}

			if (versenden.size() > 0 && isVersandAktiv()) {
				versendeMeldungen(versenden, spontanVersand);
			}
		}
	}

	private boolean isVersandAktiv() {
		return !lmsParameter.getLMSHostRechner().isEmpty();
	}

	private void versendeMeldungen(final List<TMCMeldung> versenden, final boolean spontan) throws TIC3Exception {
		if ((lmsParameter.isVersandSpontan() && spontan) || (!lmsParameter.isVersandSpontan()
				&& (System.currentTimeMillis() - letzterVersand) > lmsParameter.getVersandzyklus())) {
			versendeMeldungen(new File(bildeDateinamen()), versenden);
			letzterVersand = System.currentTimeMillis();
		}
	}

	@Override
	public void update(final ResultData[] results) {
		for (final ResultData datensatz : results) {
			if (datensatz.hasData()) {
				if (datensatz.getDataDescription().equals(paramEmpfang)) {
					debug.info("Parameterdatensatz erhalten für ",
							datensatz.getObject() + ", " + datensatz.getDataDescription().getAttributeGroup());
					lmsParameter.aktualisieren(datensatz.getData());

					synchronized (this) {
						parameterEmpfangen = true;
						this.notify();
					}
				}
			}
		}
	}

	/**
	 * Versendet eine Liste von Meldungen.
	 *
	 * @param datei
	 *            Date zum Versand der Meldungen
	 * @param neuemeldungen
	 *            zu versendende Meldungen
	 * @throws TIC3Exception
	 *             bei Ausnahmen
	 */
	public void versendeMeldungen(final File datei, final List<TMCMeldung> neuemeldungen) throws TIC3Exception {
		synchronized (runLock) {
			try {
				boolean senden = false;
				final List<TMCMeldung> versendet = new ArrayList<>();

				final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				final Document doc = factory.newDocumentBuilder().newDocument();

				// remove 'standalone=no'
				doc.setXmlStandalone(true);

				final Element root = doc.createElement(Meldungskodierer.XML_ELEMENT_TIC_ROOT);
				doc.appendChild(root);
				// <TIC version="4.26.63.1" created="2015-10-13T12:36:08.378Z"
				// culture="de-DE">
				root.setAttribute("culture", "de-DE");
				root.setAttribute("created", XMLTools.getTICDateTime(System.currentTimeMillis()));
				root.setAttribute("version", "4.26.63.1");

				for (final TMCMeldung meldung : neuemeldungen) {
					debug.info("Die TMC-Meldung: " + meldung + " wird versendet");
					try {
						meldung.datenvervollstaendigen(lmsParameter);
						final String ticId = meldung.bildeTICId();
						final ExterneMeldungsDaten md = (new TIC3Converter(null))
								.erzeugeExterneMeldungsdaten(meldung.getDavData(), ticId);
						md.erzeugeDocumentElement(doc, root);
						meldung.setLetzterVersand(System.currentTimeMillis());
						versendet.add(meldung);
						senden = true;
					} catch (final Exception e) {
						LMSTIC3LogTools.log(debug, LmsMsg.ErrorAppException, e.getMessage());
						meldung.setStatus(MeldungsStatus.FEHLERHAFT);
					}
				}
				laufendeNummerAktualisieren();

				if (senden) {
					XMLTools.schreibeXMLDatei(doc, datei);

					ftpclient.setVersandDatei(datei);
					synchronized (ftpclient) {
						ftpclient.notify();
					}
					ftpclient.interrupt();
				}
			} catch (final Exception e) {
				e.printStackTrace();
				throw new TIC3Exception(e.getMessage());
			}
		}
	}

	@Override
	public void versendet(final List<File> dateien, final LMSKommunikationsstatus status) {
		synchronized (meldungen) {
			for (final File datei : dateien) {
				if (datei.getPath().equals(LMSProperties.KEEP_ALIVE_DATEINAME)) {
					continue;
				}
				if (status == LMSKommunikationsstatus.NICHT_GESTOERT) {
					for (final TMCMeldung meldung : meldungen.values()) {
						if (!meldung.isDavDatenOK() || (meldung.getStatus() == MeldungsStatus.FEHLERHAFT)
								|| !meldung.istZeitGueltig()) {
							continue;
						}

						debug.fine("Meldung " + meldung + " erfolgreich versendet");
						meldung.setStatus(MeldungsStatus.VERSENDET);
						meldung.publiziereVersendet(dav);
						meldung.setVersendet(true);
					}
				}
				LMSTIC3LogTools.log(debug, LmsMsg.FineDateiGeloescht, datei);
				datei.delete();
			}
		}
	}

	@Override
	public void invalidObject(final DynamicObject dynamicObject) {
		debug.info("Meldung: " + dynamicObject.getPid() + " wurde invalidiert!");
		entfernenMeldung(dynamicObject);
	}
}
