/*
 * Segment 2 (KEx), SWE LMS-TIC3
 * Copyright (C) 2016 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.lmstic3.vew;

import java.text.DateFormat;
import java.text.MessageFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

/**
 * Betriebsmeldungen der SWE.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: $
 */
public final class LMTIC3Betriebsmeldungen {

	private static final String APPLIKATIONSKENNUNG = "KEx LMS-TIC3";
	private static final String ID = "KEx-ÜK-LMS";

	/**
	 * Die Betriebsmeldungen der SWE.
	 */
	public enum LMTIC3Betriebsmeldung {

		/** Kommunikationstörung ausgehend. */
		KOMMUNIKATION_AUSGEHEND_GESTOERT("KEx-ÜK-LMS01", MessageGrade.WARNING, MessageState.MESSAGE, "Der FTP-Server der LMS zur Übertragung von Verkehrsmeldungen antwortet nicht."),
		/** Kommunikation ausgehend OK. */
		KOMMUNIKATION_AUSGEHEND_OK("KEx-ÜK-LMS02", MessageGrade.INFORMATION, MessageState.GOOD_MESSAGE, "Der FTP-Server der LMS zur Übertragung von Verkehrsmeldungen antwortet wieder."),

		/** Dateifehler. */
		DATEIFEHLER("KEx-ÜK-LMS03", MessageGrade.INFORMATION, MessageState.MESSAGE, "Verarbeitung der Datei mit den Verkehrsmeldungen der LMS mit Zeitstempel {0} Uhr {1}fehlgeschlagen."),

		/** Kommunikationstörung eingehend. */
		KOMMUNIKATION_EINGEHEND_GESTOERT("KEx-ÜK-LMS04", MessageGrade.WARNING, MessageState.MESSAGE, "Der Empfang von TMC-Meldungen der Landesmeldestelle ist gestört."),
		/** Kommunikation eingehend OK. */
		KOMMUNIKATION_EINGEHEND_OK("KEx-ÜK-LMS05", MessageGrade.INFORMATION, MessageState.GOOD_MESSAGE, "Die Empfangsstörung von TMC-Meldungen der Landesmeldestelle ist beendet.");

		private final String id;
		private final MessageGrade grade;
		private final MessageState state;
		private final String message;

		LMTIC3Betriebsmeldung(final String id, final MessageGrade grade, final MessageState state,
				final String message) {
			this.id = id;
			this.grade = grade;
			this.state = state;
			this.message = message;
		}

		/**
		 * Gibt die Id der Meldung zur&uuml;ck.
		 * 
		 * @return Id
		 */
		public String getId() {
			return id;
		}

		/**
		 * Gibt den {@link MessageGrade} der Meldung zur&uuml;ck.
		 * 
		 * @return {@link MessageGrade}
		 */
		public MessageGrade getGrade() {
			return grade;
		}

		/**
		 * Gibt den {@link MessageState} der Meldung zur&uuml;ck.
		 * 
		 * @return {@link MessageState}
		 */
		public MessageState getState() {
			return state;
		}

		/**
		 * Gibt den Text der Meldung zur&uuml;ck.
		 * 
		 * @return Text
		 */
		public String getMessage() {
			return message;
		}
	}

	private static MessageSender instance;

	private LMTIC3Betriebsmeldungen() {

	}

	private static void init() {
		if (instance == null) {
			instance = MessageSender.getInstance();
			instance.setApplicationLabel(APPLIKATIONSKENNUNG);
		}
	}

	/**
	 * Versendet eine Betriebsmeldung.
	 * 
	 * @param meldung
	 *            {@link LMTIC3Betriebsmeldung}
	 */
	public static void sendeMeldung(final LMTIC3Betriebsmeldung meldung) {
		init();
		instance.sendMessage(ID, MessageType.APPLICATION_DOMAIN, "[" + meldung.getId() + "]", meldung.getGrade(),
				meldung.getState(), meldung.getMessage());
	}

	/**
	 * Sendet eine Betriebsmeldung.
	 * 
	 * @param messageId
	 *            id
	 * @param grade
	 *            {@link MessageGrade}
	 * @param state
	 *            {@link MessageState}
	 * @param message
	 *            Text
	 */
	private static void sendeMeldung(final String messageId, final MessageGrade grade, final MessageState state,
			final String message) {
		init();
		instance.sendMessage(ID, MessageType.APPLICATION_DOMAIN, "[" + messageId + "]", grade, state, message);
	}

	/**
	 * Versendet eine Dateifehler-Betriebsmeldung.
	 * 
	 * @param timestamp
	 *            Zeitstempel
	 * @param teilweise
	 *            Datei ist teilweise (<code>true</code>) oder vollständig
	 *            (<code>false</code>) defekt
	 */
	public static void sendeDateifehler(final long timestamp, final boolean teilweise) {
		final LMTIC3Betriebsmeldung dateifehler = LMTIC3Betriebsmeldung.DATEIFEHLER;

		final DateFormat df = new SimpleDateFormat("dd.MM.yyyy HH:mm");
		final String msg = MessageFormat.format(dateifehler.getMessage(), df.format(new Date(timestamp)),
				teilweise ? "teilweise " : "");
		sendeMeldung(dateifehler.getId(), dateifehler.getGrade(), dateifehler.getState(), msg);
	}
}
